"""EdgeTransformer class to convert edges to edge embeddings."""
from typing import List

import numpy as np
import pandas as pd

from .node_transformer import NodeTransformer


class EdgeTransformer:
    """EdgeTransformer class to convert edges to edge embeddings."""

    methods = {
        "Hadamard": lambda x1, x2: np.multiply(x1, x2, out=x1),
        "Sum": lambda x1, x2: np.add(x1, x2, out=x1),
        "Average": lambda x1, x2: np.divide(np.add(x1, x2, out=x1), 2, out=x1),
        "L1": lambda x1, x2: np.subtract(x1, x2, out=x1),
        "AbsoluteL1": lambda x1, x2: np.abs(np.subtract(x1, x2, out=x1), out=x1),
        "L2": lambda x1, x2: np.power(np.subtract(x1, x2, out=x1), 2, out=x1),
        "Concatenate": lambda x1, x2: np.hstack((x1, x2)),
    }

    def __init__(self, method: str = "Hadamard"):
        """Create new EdgeTransformer object.

        Parameters
        ------------------------
        method: str = "Hadamard",
            Method to use for the embedding.
            Can either be 'Hadamard', 'Sum', 'Average', 'L1', 'AbsoluteL1', 'L2' or 'Concatenate'.
        """
        if isinstance(method, str) and method not in EdgeTransformer.methods:
            raise ValueError((
                "Given method '{}' is not supported. "
                "Supported methods are {}, or alternatively a lambda."
            ).format(
                method, ", ".join(list(EdgeTransformer.methods.keys()))
            ))
        self._method = EdgeTransformer.methods[method]
        self._transformer = NodeTransformer()

    def fit(self, embedding: pd.DataFrame):
        """Fit the model.

        Parameters
        -------------------------
        embedding: pd.DataFrame,
            Embedding to use to fit the transformer.
            This is a pandas DataFrame and NOT a numpy array because we need
            to be able to remap correctly the vector embeddings in case of
            graphs that do not respect the same internal node mapping but have
            the same node set. It is possible to remap such graphs using
            Ensmallen's remap method but it may be less intuitive to users.
        """
        self._transformer.fit(embedding)

    def transform(self, sources: List[str], destinations: List[str], aligned_node_mapping: bool = False) -> np.ndarray:
        """Return embedding for given edges using provided method.

        Parameters
        --------------------------
        sources: List[str],
            List of source nodes whose embedding is to be returned.
        destinations: List[str],
            List of destination nodes whose embedding is to be returned.
        aligned_node_mapping: bool = False,
            This parameter specifies wheter the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        return self._method(
            self._transformer.transform(
                sources,
                aligned_node_mapping
            ),
            self._transformer.transform(
                destinations,
                aligned_node_mapping
            )
        )



"""LinkPredictionTransformer class to convert graphs to edge embeddings to execute link prediction."""
from typing import Tuple, Union, List
import pandas as pd
import numpy as np
from ensmallen_graph import EnsmallenGraph  # pylint: disable=no-name-in-module

from .graph_transformer import GraphTransformer


class LinkPredictionTransformer:
    """LinkPredictionTransformer class to convert graphs to edge embeddings."""

    def __init__(self, method: str = "Hadamard"):
        """Create new LinkPredictionTransformer object.

        Parameters
        ------------------------
        method: str = "hadamard",
            Method to use for the embedding.
            Can either be 'Hadamard', 'Sum', 'Average', 'L1', 'AbsoluteL1', 'L2' or 'Concatenate'.
        """
        self._transformer = GraphTransformer(method=method)

    def fit(self, embedding: pd.DataFrame):
        """Fit the model.

        Parameters
        -------------------------
        embedding: pd.DataFrame,
            Embedding to use to fit the transformer.
            This is a pandas DataFrame and NOT a numpy array because we need
            to be able to remap correctly the vector embeddings in case of
            graphs that do not respect the same internal node mapping but have
            the same node set. It is possible to remap such graphs using
            Ensmallen's remap method but it may be less intuitive to users.
        """
        self._transformer.fit(embedding)

    def transform(
        self,
        positive_graph: Union[EnsmallenGraph, np.ndarray, List[List[str]], List[List[int]]],
        negative_graph: Union[EnsmallenGraph, np.ndarray, List[List[str]], List[List[int]]],
        aligned_node_mapping: bool = False,
        random_state: int = 42
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Return edge embedding for given graph using provided method.

        Parameters
        --------------------------
        positive_graph: Union[EnsmallenGraph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an EnsmallenGraph or a list of lists of edges.
        negative_graph: Union[EnsmallenGraph, List[List[str]], List[List[int]]],
            The graph whose edges are to be embedded and labeled as positives.
            It can either be an EnsmallenGraph or a list of lists of edges.
        aligned_node_mapping: bool = False,
            This parameter specifies wheter the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.
        random_state: int = 42,
            The random state to use to shuffle the labels.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Tuple with X and y values.
        """
        positive_edge_embedding = self._transformer.transform(positive_graph, aligned_node_mapping)
        negative_edge_embedding = self._transformer.transform(negative_graph, aligned_node_mapping)
        edge_embeddings = np.vstack([
            positive_edge_embedding,
            negative_edge_embedding
        ])
        edge_labels = np.concatenate([
            np.ones(positive_edge_embedding.shape[0]),
            np.zeros(negative_edge_embedding.shape[0])
        ])
        numpy_random_state = np.random.RandomState(  # pylint: disable=no-member
            seed=random_state
        )
        indices = numpy_random_state.permutation(edge_labels.size)

        return edge_embeddings[indices], edge_labels[indices]



"""Keras Sequence object for running CBOW and SkipGram on graph walks."""
from typing import Tuple

import numpy as np  # type: ignore

from .abstract_node2vec_sequence import AbstractNode2VecSequence


class Node2VecSequence(AbstractNode2VecSequence):
    """Keras Sequence object for running CBOW and SkipGram on graph walks."""

    def __getitem__(self, idx: int) -> Tuple[Tuple[np.ndarray, np.ndarray], None]:
        """Return batch corresponding to given index.

        The return tuple of tuples is composed of an inner tuple, containing
        the words vector and the vector of vectors of the contexts.
        Depending on the order of the input_layers of the models that can
        accept these data format, one of the vectors is used as training
        input and the other one is used as the output for the NCE loss layer.

        The words vectors and contexts vectors contain numeric IDs, that
        represent the index of the words' embedding column.

        The true output value is None, since no loss function is used after
        the NCE loss, that is implemented as a layer, and this vastly improves
        the speed of the training process since it does not require to allocate
        empty vectors of considerable size for the one-hot encoding process.

        A batch returns words and contexts from:

            (number of nodes provided in a batch) *
            (number of iterations of random walks per node) *
            (walk length - window_size*2)

        different contexts.

        Parameters
        ---------------
        idx: int,
            Index corresponding to batch to be returned.

        Returns
        ---------------
        Tuple of tuples with input data.
        """
        words, contexts = self._graph.node2vec(
            self._batch_size,
            self._walk_length,
            iterations=self._iterations,
            window_size=self._window_size,
            return_weight=self._return_weight,
            explore_weight=self._explore_weight,
            change_node_type_weight=self._change_node_type_weight,
            change_edge_type_weight=self._change_edge_type_weight,
            dense_node_mapping=self._dense_node_mapping,
            max_neighbours=self._max_neighbours,
            random_state=self._random_state + idx + self.elapsed_epochs
        )

        if self._support_mirror_strategy:
            return (words.astype(float), contexts.astype(float)), None
        return (words, contexts), None



"""NodeTransformer class to convert nodes to edge embeddings."""
from typing import List, Union
import numpy as np
import pandas as pd


class NodeTransformer:
    """NodeTransformer class to convert nodes to edge embeddings."""

    def __init__(self):
        """Create new NodeTransformer object."""
        self._embedding = None
        self._embedding_numpy = None

    def fit(self, embedding: pd.DataFrame):
        """Fit the model.

        Parameters
        -------------------------
        embedding: pd.DataFrame,
            Embedding to use to fit the transformer.
            This is a pandas DataFrame and NOT a numpy array because we need
            to be able to remap correctly the vector embeddings in case of
            graphs that do not respect the same internal node mapping but have
            the same node set. It is possible to remap such graphs using
            Ensmallen's remap method but it may be less intuitive to users.
        """
        if not isinstance(embedding, pd.DataFrame):
            raise ValueError("Given embedding is not a pandas DataFrame.")
        self._embedding = embedding
        self._embedding_numpy = embedding.to_numpy()

    def transform(self, nodes: Union[List[str], List[int]], aligned_node_mapping: bool = False) -> np.ndarray:
        """Return embeddings from given node.

        Parameters
        --------------------------
        nodes: Union[List[str], List[int]],
            List of nodes whose embedding is to be returned.
            By default this should be a list of strings, if the
            aligned_node_mapping is setted, then this methods also accepts
            a list of ints.
        aligned_node_mapping: bool = False,
            This parameter specifies wheter the mapping of the embeddings nodes
            matches the internal node mapping of the given graph.
            If these two mappings do not match, the generated edge embedding
            will be meaningless.

        Raises
        --------------------------
        ValueError,
            If embedding is not fitted.

        Returns
        --------------------------
        Numpy array of embeddings.
        """
        if self._embedding is None:
            raise ValueError(
                "Transformer was not fitted yet."
            )

        if aligned_node_mapping:
            return self._embedding_numpy[nodes]

        return self._embedding.loc[nodes].to_numpy()



"""Module with embedding visualization tools."""
from multiprocessing import cpu_count
from typing import Dict, List, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from ensmallen_graph import EnsmallenGraph  # pylint: disable=no-name-in-module
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.legend_handler import HandlerBase
from sanitize_ml_labels import sanitize_ml_labels

from ..transformers import GraphTransformer, NodeTransformer


class GraphVisualizations:
    """Tools to visualize the graph embeddings."""

    DEFAULT_SCATTER_KWARGS = dict(
        s=1,
        marker=".",
        alpha=0.7
    )

    def __init__(self, method: str = "Hadamard"):
        """Create new GraphVisualizations object.

        Parameters
        -----------------------
        method: str = "Hadamard",
            Edge embedding method.
            Can either be 'Hadamard', 'Sum', 'Average', 'L1', 'AbsoluteL1', 'L2' or 'Concatenate'.
        """
        self._graph_transformer = GraphTransformer(method=method)
        self._node_transformer = NodeTransformer()
        self._node_mapping = self._node_embedding = self._edge_embedding = None
        self._method = method

    def tsne(self, X: np.ndarray, **kwargs: Dict) -> np.ndarray:
        """Return TSNE embedding of given array.

        Depending on what is available, we use tsnecuda or MulticoreTSNE.

        Parameters
        -----------------------
        X: np.ndarray,
            The data to embed.
        **kwargs: Dict,
            Parameters to pass directly to TSNE.

        Returns
        -----------------------
        The obtained TSNE embedding.
        """
        try:
            from tsnecuda import TSNE
        except ModuleNotFoundError:
            from MulticoreTSNE import MulticoreTSNE as TSNE
            if "n_jobs" not in kwargs:
                kwargs["n_jobs"] = cpu_count()
        return TSNE(**kwargs).fit_transform(X)

    def _shuffle(self, *args: List[np.ndarray]) -> List[np.ndarray]:
        """Return given arrays shuffled synchronously."""
        # Shuffling points to avoid having artificial clusters
        # caused by positions.
        index = np.arange(args[0].shape[0])
        np.random.shuffle(index)
        return [
            arg[index]
            for arg in args
        ]

    def _clear_axes(self, axes: Axes, title: str):
        """Reset the axes ticks and set the given title."""
        axes.set_xticks([])
        axes.set_xticks([], minor=True)
        axes.set_yticks([])
        axes.set_yticks([], minor=True)
        axes.set_title(title)

    def _set_legend(
        self,
        axes: Axes,
        labels: List[str],
        handles: List[HandlerBase]
    ):
        """Set the legend with the given values and handles transparency.

        Parameters
        ----------------------------
        axes: Axes,
            The axes on which to put the legend.
        labels: List[str],
            Labels to put in the legend.
        handles: List,
            Handles to display in the legend (the curresponding matplotlib
            objects).
        """
        legend = axes.legend(
            handles=handles,
            labels=sanitize_ml_labels(labels),
            loc='best'
        )
        # Setting alpha level in the legend to avoid having a transparent
        # legend scatter dots.
        for legend_handle in legend.legendHandles:
            legend_handle._legmarker.set_alpha(  # pylint: disable=protected-access
                1
            )

    def fit_transform_nodes(
        self,
        graph: EnsmallenGraph,
        embedding: pd.DataFrame,
        **kwargs: Dict
    ):
        """Executes fitting for plotting node embeddings.

        Parameters
        -------------------------
        graph: EnsmallenGraph,
            Graph from where to extract the nodes.
        embedding: pd.DataFrame,
            Embedding obtained from SkipGram, CBOW or GloVe.
        **kwargs: Dict,
            Data to pass directly to TSNE.
        """
        self._node_transformer.fit(embedding)
        self._node_embedding = self.tsne(
            self._node_transformer.transform(graph.get_node_names()),
            **kwargs
        )

    def fit_transform_edges(
        self,
        graph: EnsmallenGraph,
        embedding: np.ndarray,
        **kwargs: Dict
    ):
        """Executes fitting for plotting edge embeddings.

        Parameters
        -------------------------
        graph: EnsmallenGraph,
            Graph from where to extract the edges.
        embedding: np.ndarray,
            Embedding obtained from SkipGram, CBOW or GloVe.
        **kwargs: Dict,
            Data to pass directly to TSNE.            
        """
        self._graph_transformer.fit(embedding)
        self._edge_embedding = self.tsne(
            self._graph_transformer.transform(graph),
            **kwargs
        )

    def plot_nodes(
        self,
        figure: Figure = None,
        axes: Axes = None,
        scatter_kwargs: Dict = None,
        **kwargs: Dict
    ):
        """Plot nodes of provided graph.

        Parameters
        ------------------------------
        graph: EnsmallenGraph,
            The graph to visualize.
        figure: Figure = None,
            Figure to use to plot. If None, a new one is created using the
            provided kwargs.
        scatter_kwargs: Dict = None,
            Kwargs to pass to the scatter plot call.
        axes: Axes = None,
            Axes to use to plot. If None, a new one is created using the
            provided kwargs.

        Raises
        ------------------------------
        ValueError,
            If edge fitting was not yet executed.

        Returns
        ------------------------------
        Figure and Axis of the plot.
        """
        if self._node_embedding is None:
            raise ValueError(
                "Node fitting must be executed before plot."
            )

        if figure is None or axes is None:
            figure, axes = plt.subplots(**kwargs)

        if scatter_kwargs is None:
            scatter_kwargs = GraphVisualizations.DEFAULT_SCATTER_KWARGS

        axes.scatter(*self._node_embedding.T, **scatter_kwargs)
        self._clear_axes(axes, "Nodes embedding")
        return figure, axes

    def plot_node_types(
        self,
        graph: EnsmallenGraph,
        k: int = 10,
        figure: Figure = None,
        axes: Axes = None,
        scatter_kwargs: Dict = None,
        other_label: str = "Other",
        **kwargs
    ) -> Tuple[Figure, Axes]:
        """Plot common node types of provided graph.

        Parameters
        ------------------------------
        graph: EnsmallenGraph,
            The graph to visualize.
        k: int = 10,
            Number of node types to visualize.
        figure: Figure = None,
            Figure to use to plot. If None, a new one is created using the
            provided kwargs.
        scatter_kwargs: Dict = None,
            Kwargs to pass to the scatter plot call.
        other_label: str = "Other",
            Label to use for edges below the top k threshold.
        axes: Axes = None,
            Axes to use to plot. If None, a new one is created using the
            provided kwargs.

        Raises
        ------------------------------
        ValueError,
            If edge fitting was not yet executed.
        ValueError,
            If given k is greater than maximum supported value (10).

        Returns
        ------------------------------
        Figure and Axis of the plot.
        """
        if self._node_embedding is None:
            raise ValueError(
                "Node fitting must be executed before plot."
            )

        if k > 10:
            raise ValueError(
                "Values of k greater than 10 are not supported!"
            )

        if figure is None or axes is None:
            figure, axes = plt.subplots(**kwargs)

        if scatter_kwargs is None:
            scatter_kwargs = GraphVisualizations.DEFAULT_SCATTER_KWARGS

        top_node_types = set(list(zip(*sorted(
            graph.get_node_type_counts().items(),
            key=lambda x: x[1],
            reverse=True
        )[:k]))[0])

        node_types = graph.get_node_types()
        node_labels = graph.get_node_types_reverse_mapping()

        for i, node_type in enumerate(node_types):
            if node_type not in top_node_types:
                node_types[i] = len(top_node_types)

        for node_type in range(graph.get_node_types_number()):
            if node_type not in top_node_types:
                node_labels[node_type] = other_label

        node_embeddding, node_types = self._shuffle(
            self._node_embedding,
            node_types
        )

        scatter = axes.scatter(
            *node_embeddding.T,
            c=node_types,
            cmap=plt.get_cmap("tab10"),
            **scatter_kwargs,
        )

        self._set_legend(
            axes,
            node_labels,
            scatter.legend_elements()[0]
        )
        self._clear_axes(axes, "Nodes types")
        return figure, axes

    def plot_node_degrees(
        self,
        graph: EnsmallenGraph,
        figure: Figure = None,
        axes: Axes = None,
        scatter_kwargs: Dict = None,
        **kwargs: Dict
    ):
        """Plot node degrees heatmap.

        Parameters
        ------------------------------
        graph: EnsmallenGraph,
            The graph to visualize.
        figure: Figure = None,
            Figure to use to plot. If None, a new one is created using the
            provided kwargs.
        scatter_kwargs: Dict = None,
            Kwargs to pass to the scatter plot call.
        axes: Axes = None,
            Axes to use to plot. If None, a new one is created using the
            provided kwargs.

        Raises
        ------------------------------
        ValueError,
            If edge fitting was not yet executed.

        Returns
        ------------------------------
        Figure and Axis of the plot.
        """
        if self._node_embedding is None:
            raise ValueError(
                "Node fitting must be executed before plot."
            )

        if figure is None or axes is None:
            figure, axes = plt.subplots(**kwargs)

        if scatter_kwargs is None:
            scatter_kwargs = GraphVisualizations.DEFAULT_SCATTER_KWARGS

        degrees = graph.degrees()
        two_median = np.median(degrees)*3
        degrees[degrees > two_median] = min(two_median, degrees.max())

        node_embeddding, degrees = self._shuffle(self._node_embedding, degrees)

        scatter = axes.scatter(
            *node_embeddding.T,
            c=degrees,
            cmap=plt.cm.get_cmap('RdYlBu'),
            **scatter_kwargs,
        )
        color_bar = figure.colorbar(scatter, ax=axes)
        color_bar.set_alpha(1)
        color_bar.draw_all()
        self._clear_axes(axes, "Nodes degrees")
        return figure, axes

    def plot_edge_types(
        self,
        graph: EnsmallenGraph,
        k: int = 10,
        figure: Figure = None,
        axes: Axes = None,
        scatter_kwargs: Dict = None,
        other_label: str = "Other",
        **kwargs: Dict
    ):
        """Plot common edge types of provided graph.

        Parameters
        ------------------------------
        graph: EnsmallenGraph,
            The graph to visualize.
        k: int = 10,
            Number of edge types to visualize.
        figure: Figure = None,
            Figure to use to plot. If None, a new one is created using the
            provided kwargs.
        scatter_kwargs: Dict = None,
            Kwargs to pass to the scatter plot call.
        other_label: str = "Other",
            Label to use for edges below the top k threshold.
        axes: Axes = None,
            Axes to use to plot. If None, a new one is created using the
            provided kwargs.

        Raises
        ------------------------------
        ValueError,
            If edge fitting was not yet executed.
        ValueError,
            If given k is greater than maximum supported value (10).

        Returns
        ------------------------------
        Figure and Axis of the plot.
        """
        if self._edge_embedding is None:
            raise ValueError(
                "Edge fitting must be executed before plot."
            )

        if k > 10:
            raise ValueError(
                "Values of k greater than 10 are not supported!"
            )

        if figure is None or axes is None:
            figure, axes = plt.subplots(**kwargs)

        if scatter_kwargs is None:
            scatter_kwargs = GraphVisualizations.DEFAULT_SCATTER_KWARGS

        top_edge_types = set(list(zip(*sorted(
            graph.get_edge_type_counts().items(),
            key=lambda x: x[1],
            reverse=True
        )[:k]))[0])

        edge_types = graph.get_edge_types()
        edge_labels = graph.get_edge_type_names()

        for i, edge_type in enumerate(edge_types):
            if edge_type not in top_edge_types:
                edge_types[i] = len(top_edge_types)

        for edge_type in range(graph.get_edge_types_number()):
            if edge_type not in top_edge_types:
                edge_labels[edge_type] = other_label

        edge_tsne, edge_types = self._shuffle(self._edge_embedding, edge_types)

        scatter = axes.scatter(
            *edge_tsne.T,
            c=edge_types,
            cmap=plt.get_cmap("tab10"),
            **scatter_kwargs,
        )

        self._set_legend(
            axes,
            edge_labels,
            scatter.legend_elements()[0]
        )
        self._clear_axes(axes, "Edge types")
        return figure, axes

    def plot_edges(
        self,
        figure: Figure = None,
        axes: Axes = None,
        scatter_kwargs: Dict = None,
        **kwargs: Dict
    ):
        """Plot edge embedding of provided graph.

        Parameters
        ------------------------------
        figure: Figure = None,
            Figure to use to plot. If None, a new one is created using the
            provided kwargs.
        scatter_kwargs: Dict = None,
            Kwargs to pass to the scatter plot call.
        axes: Axes = None,
            Axes to use to plot. If None, a new one is created using the
            provided kwargs.

        Raises
        ------------------------------
        ValueError,
            If edge fitting was not yet executed.

        Returns
        ------------------------------
        Figure and Axis of the plot.
        """
        if self._edge_embedding is None:
            raise ValueError(
                "Edge fitting must be executed before plot."
            )

        if figure is None or axes is None:
            figure, axes = plt.subplots(**kwargs)

        if scatter_kwargs is None:
            scatter_kwargs = GraphVisualizations.DEFAULT_SCATTER_KWARGS

        axes.scatter(*self._edge_embedding.T, **scatter_kwargs)
        self._clear_axes(
            axes,
            "Edge embeddings with method {method}".format(
                method=self._method
            )
        )
        return figure, axes

    def plot_edge_weights(
        self,
        graph: EnsmallenGraph,
        figure: Figure = None,
        axes: Axes = None,
        scatter_kwargs: Dict = None,
        **kwargs: Dict
    ):
        """Plot common edge types of provided graph.

        Parameters
        ------------------------------
        graph: EnsmallenGraph,
            The graph to visualize.
        figure: Figure = None,
            Figure to use to plot. If None, a new one is created using the
            provided kwargs.
        scatter_kwargs: Dict = None,
            Kwargs to pass to the scatter plot call.
        axes: Axes = None,
            Axes to use to plot. If None, a new one is created using the
            provided kwargs.

        Raises
        ------------------------------
        ValueError,
            If edge fitting was not yet executed.

        Returns
        ------------------------------
        Figure and Axis of the plot.
        """
        if self._edge_embedding is None:
            raise ValueError(
                "Edge fitting must be executed before plot."
            )

        if figure is None or axes is None:
            figure, axes = plt.subplots(**kwargs)

        if scatter_kwargs is None:
            scatter_kwargs = GraphVisualizations.DEFAULT_SCATTER_KWARGS

        edge_embedding, weights = self._shuffle(
            self._edge_embedding,
            graph.get_weights()
        )

        scatter = axes.scatter(
            *edge_embedding.T,
            c=weights,
            cmap=plt.cm.get_cmap('RdYlBu'),
            **scatter_kwargs,
        )
        color_bar = figure.colorbar(scatter, ax=axes)
        color_bar.set_alpha(1)
        color_bar.draw_all()
        self._clear_axes(axes, "Edge weights")
        return figure, axes



"""GloVe model for graph and words embedding."""
from typing import Union

import tensorflow as tf
from tensorflow.keras import backend as K  # pylint: disable=import-error
from tensorflow.keras.layers import Add, Dot, Embedding, Flatten, Input  # pylint: disable=import-error
from tensorflow.keras.models import Model   # pylint: disable=import-error
from tensorflow.keras.optimizers import Optimizer   # pylint: disable=import-error

from .embedder import Embedder


class GloVe(Embedder):
    """GloVe model for graph and words embedding.

    The GloVe model for graoh embedding receives two words and is asked to
    predict its cooccurrence probability.
    """

    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        optimizer: Union[str, Optimizer] = "nadam",
        alpha: float = 0.75,
        shared_embedding_layers: bool = False
    ):
        """Create new GloVe-based Embedder object.

        Parameters
        ----------------------------
        vocabulary_size: int,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
        embedding_size: int,
            Dimension of the embedding.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.
        alpha: float = 0.75,
            Alpha to use for the function.
        shared_embedding_layers: bool = False,
            Whether to share the embedding layers weights
            for the center words and the contexts.
            This will make the GloVe model more practical as it halves the number
            of parameters of the model, but it is still to be studied properly.
        """
        self._alpha = alpha
        self._shared_embedding_layers = shared_embedding_layers
        super().__init__(
            vocabulary_size=vocabulary_size,
            embedding_size=embedding_size,
            optimizer=optimizer
        )

    def _glove_loss(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> float:
        """Compute the glove loss function.

        Parameters
        ---------------------------
        y_true: tf.Tensor,
            The true values Tensor for this batch.
        y_pred: tf.Tensor,
            The predicted values Tensor for this batch.

        Returns
        ---------------------------
        Loss function score related to this batch.
        """
        return K.sum(
            K.pow(K.clip(y_true, 0.0, 1.0), self._alpha) *
            K.square(y_pred - K.log(y_true)),
            axis=-1
        )

    def _build_model(self):
        """Create new Glove model."""
        # Creating the input layers
        input_layers = [
            Input((1,), name=Embedder.EMBEDDING_LAYER_NAME),
            Input((1,))
        ]

        # Creating the embedding layer(s)
        if self._shared_embedding_layers:
            embedding_layer = Embedding(
                self._vocabulary_size,
                self._embedding_size,
                input_length=1
            )
            embedding_layers = [
                embedding_layer(input_layer)
                for input_layer in input_layers
            ]
        else:
            embedding_layers = [
                Embedding(
                    self._vocabulary_size,
                    self._embedding_size,
                    input_length=1
                )(input_layer)
                for input_layer in input_layers
            ]

        # Creating the dot product of the embedding layers
        dot_product_layer = Dot(axes=2)(embedding_layers)

        # Creating the biases layer
        biases = [
            Embedding(
                self._vocabulary_size,
                1,
                input_length=1
            )(input_layer)
            for input_layer in input_layers
        ]

        # Concatenating with an add the three layers
        prediction = Flatten()(Add()([dot_product_layer, *biases]))

        # Creating the model
        glove = Model(
            inputs=input_layers,
            outputs=prediction,
            name="GloVe"
        )
        glove.compile(
            loss=self._glove_loss,
            optimizer=self._optimizer
        )

        return glove



"""Abstract Keras Sequence object for running models on graph walks."""
from typing import Dict

from ensmallen_graph import EnsmallenGraph  # pylint: disable=no-name-in-module
from .abstract_sequence import AbstractSequence


class AbstractNode2VecSequence(AbstractSequence):
    """Abstract Keras Sequence object for running models on graph walks."""

    def __init__(
        self,
        graph: EnsmallenGraph,
        walk_length: int,
        batch_size: int,
        iterations: int = 1,
        window_size: int = 4,
        return_weight: float = 1.0,
        explore_weight: float = 1.0,
        change_node_type_weight: float = 1.0,
        change_edge_type_weight: float = 1.0,
        max_neighbours: int = None,
        elapsed_epochs: int = 0,
        support_mirror_strategy: bool = False,
        seed: int = 42,
        dense_node_mapping: Dict[int, int] = None
    ):
        """Create new Node2Vec Sequence object.

        Parameters
        -----------------------------
        graph: EnsmallenGraph,
            The graph from from where to extract the walks.
        walk_length: int,
            Maximal length of the walks.
            In directed graphs, when traps are present, walks may be shorter.
        batch_size: int,
            Number of nodes to include in a single batch.
        iterations: int = 1,
            Number of iterations of the single walks.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        shuffle: bool = True,
            Whether to shuffle the vectors.
        return_weight: float = 1.0,
            Weight on the probability of returning to the same node the walk just came from
            Having this higher tends the walks to be
            more like a Breadth-First Search.
            Having this very high  (> 2) makes search very local.
            Equal to the inverse of p in the Node2Vec paper.
        explore_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node
            to the one we're coming from in the random walk
            Having this higher tends the walks to be
            more like a Depth-First Search.
            Having this very high makes search more outward.
            Having this very low makes search very local.
            Equal to the inverse of q in the Node2Vec paper.
        change_node_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor node of a
            different type than the previous node. This only applies to
            colored graphs, otherwise it has no impact.
        change_edge_type_weight: float = 1.0,
            Weight on the probability of visiting a neighbor edge of a
            different type than the previous edge. This only applies to
            multigraphs, otherwise it has no impact.
        max_neighbours: int = None,
            Number of maximum neighbours to consider when using approximated walks.
            By default, None, we execute exact random walks.
            This is mainly useful for graphs containing nodes with extremely high degrees.
        elapsed_epochs: int = 0,
            Number of elapsed epochs to init state of generator.
        support_mirror_strategy: bool = False,
            Wethever to patch support for mirror strategy.
            At the time of writing, TensorFlow's MirrorStrategy does not support
            input values different from floats, therefore to support it we need
            to convert the unsigned int 32 values that represent the indices of
            the embedding layers we receive from Ensmallen to floats.
            This will generally slow down performance, but in the context of
            exploiting multiple GPUs it may be unnoticeable.
        dense_node_mapping: Dict[int, int] = None,
            Mapping to use for converting sparse walk space into a dense space.
            This object can be created using the method (available from the
            graph object created using EnsmallenGraph)
            called `get_dense_node_mapping` that returns a mapping from
            the non trap nodes (those from where a walk could start) and
            maps these nodes into a dense range of values.
        """
        self._graph = graph
        self._walk_length = walk_length
        self._iterations = iterations
        self._return_weight = return_weight
        self._explore_weight = explore_weight
        self._max_neighbours = max_neighbours
        self._change_node_type_weight = change_node_type_weight
        self._change_edge_type_weight = change_edge_type_weight
        self._dense_node_mapping = dense_node_mapping

        super().__init__(
            batch_size=batch_size,
            sample_number=self._graph.get_unique_sources_number(),
            window_size=window_size,
            elapsed_epochs=elapsed_epochs,
            support_mirror_strategy=support_mirror_strategy,
            random_state=seed
        )



"""Abstract Keras Model object for embedding models."""
from typing import Union, List

import numpy as np
import pandas as pd
from tensorflow.keras.models import Model   # pylint: disable=import-error
from tensorflow.keras.optimizers import Optimizer   # pylint: disable=import-error


class Embedder:
    """Abstract Keras Model object for embedding models."""

    EMBEDDING_LAYER_NAME = "words_embedding"

    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        optimizer: Union[str, Optimizer] = "nadam"
    ):
        """Create new Embedder object.

        Parameters
        ----------------------------------
        vocabulary_size: int,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
        embedding_size: int,
            Dimension of the embedding.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.

        Raises
        -----------------------------------
        ValueError,
            When the given vocabulary size is not a strictly positive integer.
        ValueError,
            When the given embedding size is not a strictly positive integer.
        """
        if not isinstance(vocabulary_size, int) or vocabulary_size < 1:
            raise ValueError((
                "The given vocabulary size ({}) "
                "is not a strictly positive integer."
            ).format(
                vocabulary_size
            ))
        if not isinstance(embedding_size, int) or embedding_size < 1:
            raise ValueError((
                "The given embedding size ({}) "
                "is not a strictly positive integer."
            ).format(
                embedding_size
            ))
        self._vocabulary_size = vocabulary_size
        self._embedding_size = embedding_size
        self._optimizer = optimizer
        self._model = self._build_model()

    def _build_model(self) -> Model:
        """Build new model for embedding."""
        raise NotImplementedError(
            "The method _build_model must be implemented in the child classes."
        )

    def summary(self):
        """Print model summary."""
        self._model.summary()

    @property
    def embedding(self) -> np.ndarray:
        """Return model embeddings."""
        for layer, weights in zip(self._model.layers, self._model.weights):
            if layer.name == Embedder.EMBEDDING_LAYER_NAME:
                return weights.numpy()
        return None

    def get_embedding_dataframe(self, term_names: List[str]) -> pd.DataFrame:
        """Return terms embedding using given index names.

        Parameters
        -----------------------------
        path: str,
            Save embedding as csv to given path.
        term_names: List[str],
            List of terms to be used as index names.
        """
        return pd.DataFrame(
            self.embedding,
            index=term_names
        )

    def save_embedding(self, path: str, term_names: List[str]):
        """Save terms embedding using given index names.

        Parameters
        -----------------------------
        path: str,
            Save embedding as csv to given path.
        term_names: List[str],
            List of terms to be used as index names.
        """
        self.get_embedding_dataframe(term_names).to_csv(path, header=False)

    @property
    def name(self) -> str:
        """Return model name."""
        return self._model.name

    def save_weights(self, path: str):
        """Save model weights to given path.

        Parameters
        ---------------------------
        path: str,
            Path where to save model weights.
        """
        self._model.save_weights(path)

    def load_weights(self, path: str):
        """Load model weights from given path.

        Parameters
        ---------------------------
        path: str,
            Path from where to load model weights.
        """
        self._model.load_weights(path)

    def fit(self, *args, **kwargs) -> pd.DataFrame:
        """Return pandas dataframe with training history."""
        return pd.DataFrame(self._model.fit(*args, **kwargs).history)



"""Abstract class for graph embedding models."""
from typing import Union, Tuple

from tensorflow.keras import backend as K   # pylint: disable=import-error
from tensorflow.keras.layers import Embedding, Input, Lambda, Layer, Flatten   # pylint: disable=import-error
from tensorflow.keras.models import Model   # pylint: disable=import-error
from tensorflow.keras.optimizers import Optimizer   # pylint: disable=import-error

from .embedder import Embedder
from .layers import NoiseContrastiveEstimation


class Node2Vec(Embedder):
    """Abstract class for graph embedding models."""

    def __init__(
        self,
        vocabulary_size: int,
        embedding_size: int,
        model_name: str,
        optimizer: Union[str, Optimizer] = "nadam",
        window_size: int = 4,
        negative_samples: int = 10
    ):
        """Create new Graph Embedder model.

        Parameters
        -------------------------------------------
        vocabulary_size: int,
            Number of terms to embed.
            In a graph this is the number of nodes, while in a text is the
            number of the unique words.
        embedding_size: int,
            Dimension of the embedding.
        model_name: str,
            Name of the model.
        optimizer: Union[str, Optimizer] = "nadam",
            The optimizer to be used during the training of the model.
        window_size: int = 4,
            Window size for the local context.
            On the borders the window size is trimmed.
        negative_samples: int,
            The number of negative classes to randomly sample per batch.
            This single sample of negative classes is evaluated for each element in the batch.
        """
        self._model_name = model_name
        self._window_size = window_size
        self._negative_samples = negative_samples
        super().__init__(
            vocabulary_size=vocabulary_size,
            embedding_size=embedding_size,
            optimizer=optimizer
        )

    def _get_true_input_length(self) -> int:
        """Return length of true input layer."""
        raise NotImplementedError((
            "The method '_get_true_input_length' "
            "must be implemented in child class."
        ))

    def _get_true_output_length(self) -> int:
        """Return length of true output layer."""
        raise NotImplementedError((
            "The method '_get_true_output_length' "
            "must be implemented in child class."
        ))

    def _sort_input_layers(
        self,
        true_input_layer: Layer,
        true_output_layer: Layer
    ) -> Tuple[Layer, Layer]:
        """Return input layers for training with the same input sequence.

        Parameters
        ----------------------------
        true_input_layer: Layer,
            The input layer that will contain the true input.
        true_output_layer: Layer,
            The input layer that will contain the true output.

        Returns
        ----------------------------
        Return tuple with the tuple of layers.
        """
        raise NotImplementedError((
            "The method '_sort_input_layers' "
            "must be implemented in child class."
        ))

    def _build_model(self):
        """Return Node2Vec model."""
        # Creating the inputs layers
        true_input_layer = Input(
            (self._get_true_input_length(), ),
            name=Embedder.EMBEDDING_LAYER_NAME
        )
        true_output_layer = Input(
            (self._get_true_output_length(), ),
        )

        # Creating the embedding layer for the contexts
        embedding = Embedding(
            input_dim=self._vocabulary_size,
            output_dim=self._embedding_size,
            input_length=self._get_true_input_length()
        )(true_input_layer)

        # If there is more than one value per single sample
        # as there is for instance in CBOW-like models
        if self._get_true_input_length() > 1:
            # Computing mean of the embedding of all the contexts
            mean_embedding = Lambda(
                lambda x: K.mean(x, axis=1),
                output_shape=(self._embedding_size,)
            )(embedding)
        else:
            # Otherwise we passthrough the previous result with a simple flatten.
            mean_embedding = Flatten()(embedding)

        # Adding layer that also executes the loss function
        nce_loss = NoiseContrastiveEstimation(
            vocabulary_size=self._vocabulary_size,
            embedding_size=self._embedding_size,
            negative_samples=self._negative_samples,
            positive_samples=self._get_true_output_length()
        )((mean_embedding, true_output_layer))

        # Creating the actual model
        model = Model(
            inputs=self._sort_input_layers(
                true_input_layer,
                true_output_layer
            ),
            outputs=nce_loss,
            name=self._model_name
        )

        # No loss function is needed because it is already executed in
        # the NCE loss layer.
        model.compile(
            optimizer=self._optimizer
        )
        return model



"""Module with Keras Sequences."""
from .node2vec_sequence import Node2VecSequence
from .link_prediction_sequence import LinkPredictionSequence
from .word2vec import Word2VecSequence

__all__ = [
    "Node2VecSequence",
    "LinkPredictionSequence",
    "Word2VecSequence"
]
